;; Generated package description from notmuch-indicator.el  -*- no-byte-compile: t -*-
(define-package "notmuch-indicator" "0.2.0.0.20221016.93535" "Display mode line indicator with notmuch-count(1) output" '((emacs "27.1")) :commit "65931b3caaef281ac0d7de3d5c1f09af07a6e5cb" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("notmuch-indicator Development" . "~protesilaos/notmuch-indicator@lists.sr.ht") :keywords '("convenience" "mail") :url "https://git.sr.ht/~protesilaos/notmuch-indicator")
